program AccessCompact;

// ***************************************************************************
//
// AccessCompact compacts and repairs Access 97 and Access 2000 databases.
//
// Author: David Simpson (drs@ihug.com.au), 19 Feb 2000
//
// Minor changes: Bob Wasaff (bwasaff@suscom.net), 29 Sep 00 2000
//                David Simpson, 30 Sep 2000
// Minor changes: Jobst Hensiek ( jobst@hensiek.de) 21. Nov.2000)
//                - Check for enough disk space!
//                - Added detection of the intalled Access # See DAO
//                - Detect the version of mdac / msjro.dll
//                - for older MDACs with ACCESS 97 switched to DAO for repair
// ***************************************************************************

{$APPTYPE CONSOLE}
{$DEFINE GERMAN}
{.$DEFINE DEBUG}
uses
  SysUtils,
  windows,
  Registry,
  ActiveX,
  JRO_TLB,
  DAO_TLB;

// derived from dao350.dll. Rename TField to T_FIELD !!!

const
{$IFDEF DEBUG}
  sDebugPath = 'D:\Balti\NORDWIND.MDB';
{$ENDIF}
  CRLF = #13#10;
  nDiskSafety   = 1000000; // at least 1 Meg spare room on current drive
  sMdacMajor = '2.';
  sProvider  = 'Provider=Microsoft.Jet.OLEDB.4.0;Data Source=';
  sErrParm   = 'Error: Unknown parameter ';
  jro_key    = 'TypeLib\{AC3B8B4C-B6CA-11D1-9F31-00C04FC29D52}'; // check on your Systems please ...
  copyright  = 'Authors: David Simpson (drs@ihug.com.au)' + CRLF +
    '         Bob Wasaff (bwasaff@suscom.net)' + CRLF +
    '         Jobst Hensiek (jobst@hensiek.de)' + CRLF + CRLF +
    'No royalties, no warranty.  Use at your own risk.  Open Source.' + CRLF +
    'MS Access is a registered trademark of Microsoft Corp.';

var
  bUseDao: Boolean;
  nAccessVersion: integer;
  sjroVersion: string;
  Param1, Param2: string;

function CheckFreeSpace(APath: string): boolean;
var
  SR: TSearchRec;
  nSizeNeeded,
    nSizeFree: Int64;
begin
  if not FileExists(APath) then
    begin
      Writeln('Error: File ', APath, ' not found.');
      Result := false;
      exit;
    end;
   try
    nSizeNeeded := nDiskSafety;
    nSizeFree := DiskFree(ORD(APath[1]) - 64); // free diskspace found
    if FindFirst(APath, faAnyFile, SR) = 0 then
      nSizeNeeded := nSizeNeeded + SR.Size; // required space incl. safety
    if nSizeNeeded < nSizeFree then
      Result := true
    else
      Result := false;
    Writeln(Format('Required disk space: %3.0n KB.', [nSizeNeeded / 1000]));
    Writeln(Format('Found disk space:    %3.0n KB.', [nSizeFree / 1000]));
  finally
    windows.FindClose(SR.FindHandle); // Tell about free space
  end;
end;

function checkJROVer(sNum: string): string;
var
  Reg: Tregistry;
begin
  Result := sMdacMajor;
  Reg := Tregistry.create;
  try
    Reg.RootKey := HKEY_CLASSES_ROOT;
    if Reg.openKey(jro_key + '\' + sMdacMajor + trim(sNum), false) then
      Result := Result + sNum
    else
      Result := '';
  finally
    Reg.Free;
  end;
end;

Procedure writeMdacError;
begin
   Writeln('Error: This utility requires MDAC Version 2.1 or higher.');
   Writeln('You should update to MDAC 2.1 or higher.');
   Writeln('-> Download from: http://www.microsoft.com/data <-');
end;

function CheckRegistry(bCheckAccess: Boolean): integer;
var
  Reg: Tregistry;
begin
  // check version of installed mdac_typ / msjro.dll
  Result := 0;
  sjroVersion := sMdacMajor;
  bUseDao := false;
  if checkJROVer('6') <> '' then
    sjroVersion := sjroVersion + '6'
  else if checkJROVer('5') <> '' then
    sjroVersion := sjroVersion + '5'
  else if checkJROVer('1') <> '' then
    begin
      sjroVersion := sjroVersion + '1';
      bUseDao := True;
    end
  else
    begin
      writeMdacError;
      Result := -1;
      exit;
    end;
  if bCheckAccess then
    begin
      Reg := TRegistry.Create;
      try
        with Reg do
          begin
            RootKey := HKEY_CLASSES_ROOT;
            OpenKey('Access.Application\CurVer', false);
            // sorry for line below
            Result := StrtoInt(Trim(ReadString(''))[Length(Trim(ReadString('')))]);
          end;
      finally
        Reg.Free;
      end;
      if ((Result = 9) and bUseDao) then
        begin
          bUseDao := false;
          writeMdacError;
        end
      else if not Result in [8, 9] then // check for Access 97(8) or 2000(9)
        Result := -1;
    end;
end;

// compact & repair ....
procedure CompactDB(const APath: string; AAccessVersion: integer);
var
  MyJetEngine: JetEngine;
  MyDaoEngine: DBEngine;
  strDataSource,
    strDataDest,
    strDataDestName: string;
begin
  // init COM-lib
  CoInitialize(nil);

  Writeln(CRLF + Format('Compacting started at %s: ...', [dateTimeToStr(now)]) + CRLF);

  // delete any previous temporary mdb file
  strDataDestName := ExtractFilePath(APath) + 'AccessCompact.mdb';
  if FileExists(StrDataDestName) then
    begin
      SysUtils.DeleteFile(strDataDestName);
      Writeln('Previous temporary file ', strDataDestName, ' deleted.');
    end;

  // the old JetEngine (mdac < 2.5) can not do a repair !! So let's do it here!
  if bUseDao then
    begin
      myDaoEngine := CoDbEngine.Create();
      try
{$IFDEF DEBUG}
  Writeln(Format('DAO-VERSION : %s.', [myDaoEngine.Version]));
{$ENDIF}
       myDaoEngine.RepairDatabase(APath);
       Writeln(APath, ' repaired.');
       myDaoEngine := nil;
      except
       on E: Exception do Writeln('Error: ', E.Message);
      end;
    end;

  strDataSource := sProvider + APath + ';';
  strDataDest :=   sProvider + strDataDestName + ';';
  // default to Access 2000 Format unless 97 Format is specified
  if ((AAccessVersion = 8) or (AAccessVersion = 80 )) then
    // Use Engine Type 4 for Access 97 db
    strDataDest := strDataDest + 'Jet OLEDB:Engine Type=4'
  else if ((AAccessVersion = 9) or (AAccessVersion = 90 )) then
    // Use Engine Type 5 for Access 2000 db
    strDataDest := strDataDest + 'Jet OLEDB:Engine Type=5';

  MyJetEngine := CoJetEngine.Create();
  try
    MyJetEngine.CompactDatabase(strDataSource, strDataDest);
    if bUseDao then
      Writeln(APath, ' compacted into ', strDataDestName, '.')
    else
      Writeln(APath, ' compacted and repaired into ', strDataDestName, '.');
    MyJetEngine := nil;
    if SysUtils.DeleteFile(APath) then
      begin
        Writeln('Old ', APath, ' deleted.');
        if RenameFile(strDataDestName, APath) then
          Writeln(strDataDestName, ' renamed ', APath, '.')
        else
          Writeln('Error: ', strDataDestName, ' could not be renamed ', APath, '.');
      end
    else
      Writeln('Error: ', APath, ' could not be deleted.');
  except
    on E: Exception do Writeln('Error: ', E.Message);
  end;
  Writeln(CRLF + Format('            ... finished at %s.', [dateTimeToStr(now)]));
  // uninit COM-lib
  CoUninitialize;
end;

procedure WriteHelp;
begin
  Writeln('AccessCompact repairs and compacts MS Access 97 and MS Access 2000');
  Writeln('databases.  Compacting also rebuilds the indexes.  This utility requires');
  Writeln('MDAC Version 2.1 or higher.  You can upgrade to MDAC 2.5 or later from');
  Writeln('http://www.microsoft.com/data.', CRLF);
  Writeln('ACCESSCOMPACT [[Drive:][Path]filename|[/,-][MD|AM|DS]] [/,-]2000|97|CA] [>File.log]', CRLF);
  Writeln('-First Parameter:');
  Writeln('  [drive:][path]filename Mandatory. Specifies database to repair & compact.');
  Writeln('  /MD   : Reports MDAC version only.');
  Writeln('  /AM   : Reports highest installed MS Access and MDAC version only.');
  Writeln('  /DS   : Reports free diskspace on current drive.', CRLF);
  Writeln('-Second Parameter:');
  Writeln('  /2000 : Force database type to 2000. This is the default. *)');
  Writeln('  /97   : Force database type to 97.');
  Writeln('  /CA   : Detect installed MS Access and use corresponding database type.', CRLF);
  Writeln('  [> File.log] directs console output into a log-file.' + CRLF);
  Writeln('*) Defaults back to Access 97 with MDAC below 2.5 and no Access 2000 installed.'+CRLF);
  write(' -- more --');readln;
  Writeln(copyright);
end;

procedure WriteMDACversion;
begin
  nAccessVersion := CheckRegistry(false);
  Writeln('MDAC version ' + sjroVersion + ' detected.');
end;

procedure WriteAccessVersion;
begin
  nAccessVersion := CheckRegistry(true);
  if nAccessVersion = 8 then
    Writeln(Format('MS Access 97 with MDAC version %s detected.', [sjroVersion]))
  else if nAccessVersion = 9 then
    Writeln(Format('MS Access 2000 with MDAC version %s detected.', [sjroVersion]))
  else
    Writeln(Format('Error: MS Access not installed, MDAC version %s detected.', [sjroVersion]));
end;

procedure WriteDiskSpace;
begin
  Writeln(Format('Free disk space: %n kBytes.', [DiskFree(ORD(GetCurrentDir[1]) - 64) / 1000.0]));
end;

function CheckAccessVersion: boolean;
begin
  nAccessVersion := CheckRegistry(true);
  if nAccessVersion < 0 then
    begin
      Writeln('Error: Neither MS Access 97 nor MS Access 2000 is installed.');
      Result := false;
    end
  else
    begin
      case nAccessVersion of
        0: Writeln(' MDAC:[#' + sjroVersion + '] detected.');
        8: Writeln(Format('MS Access 97 with MDAC:[#%s] detected.', [sjroVersion]));
        9: Writeln(Format('MS Access 2000 with MDAC:[#%s] detected.', [sjroVersion])); // Access 2000
        80: Writeln(Format('Forced to MS Access 97 database with MDAC:[#%s] detected.', [sjroVersion]));
        90: Writeln(Format('Forced to MS Access 2000 database. Detected MDAC:[#%s].', [sjroVersion])); // Access 2000
      end;
      Result := true;
    end;
end;

// procedure  main. Here it starts
begin
  Writeln(CRLF);
{$IFDEF DEBUG}
  Writeln(Format('ParamCount %d, Paramstr(2)%s', [paramcount, paramstr(2)]));
{$ENDIF}

  case paramcount of
    0: Writeln('Use AccessCompact /? to get help.');
    1:
      begin
        Param1 := upperCase(Paramstr(1));
        if (Param1 = '/?') or (Param1 = '-?') or (Param1 = '?') then
          WriteHelp
        else if (Param1 = '/MD') or (Param1 = '-MD') then
          WriteMDACversion
        else if (Param1 = '/AM') or (Param1 = '-AM') then
          WriteAccessVersion
        else if (Param1 = '/DS') or (Param1 = '-DS') then
          WriteDiskSpace
        else if Copy(Param1, Length(Param1) - 3, 4) = '.MDB' then
          begin
            nAccessVersion := CheckRegistry(true);
{$IFDEF GERMAN}
            decimalseparator := '.';
{$ENDIF}
            if ((nAccessVersion < 9) and (StrToFloat(trim(sjroVersion)) < 2.5)) then begin
               nAccessVersion := 80;
               Writeln(#7+'*) Defaults to Access 97 with MDAC # below 2.5 and no Access 2000 installed.'+CRLF);
            end else
               nAccessVersion := 90;
{$IFDEF GERMAN}
            decimalseparator := ',';
{$ENDIF}
            if CheckFreeSpace(Param1) then
                       CompactDB(ParamStr(1), nAccessVersion);
          end
        else
          Writeln(sErrParm, Param1, '.');
      end;
    2:
      begin
        Param1 := upperCase(Paramstr(1));
        Param2 := upperCase(Paramstr(2));
        if Copy(Param1, Length(Param1) - 3, 4) <> '.MDB' then
          begin
            Writeln(sErrParm, Param1, '.');
            exit;
          end;
        if (Param2 = '/CA') or (Param2 = '-CA') then
          begin
            if CheckAccessVersion then
              if CheckFreeSpace(Param1) then
                 CompactDB(ParamStr(1), nAccessVersion);
           end
        else if (Param2 = '/97') or (Param2 = '-97') then
          begin
            nAccessVersion := CheckRegistry(false);
            nAccessVersion := 80;
            if CheckFreeSpace(Param1) then
               CompactDB(ParamStr(1), nAccessVersion);
          end
        else if (Param2 = '/2000') or (Param2 = '-2000') then
          begin
            nAccessVersion := CheckRegistry(true);
{$IFDEF GERMAN}
            decimalseparator := '.';
{$ENDIF}
            if ((nAccessVersion < 9) and (StrToFloat(sjroVersion) < 2.5)) then
               nAccessVersion := 80
            else
               nAccessVersion := 90;
{$IFDEF GERMAN}
            decimalseparator := ',';
{$ENDIF}
            if CheckFreeSpace(Param1) then
               CompactDB(ParamStr(1), nAccessVersion);
          end
        else
          Writeln(sErrParm, Param2, '.');
      end;
  end;
end.

